@@ -71,7 +71,7 @@ def resource(public_id, **options):
     resource_type = options.pop("resource_type", "image")
     type = options.pop("type", "upload")
     uri = ["resources", resource_type, type, public_id]
-    return call_api("get", uri, only(options, "exif", "faces", "colors", "image_metadata", "pages", "phash", "max_results"), **options)
+    return call_api("get", uri, only(options, "exif", "faces", "colors", "image_metadata", "pages", "phash", "coordinates", "max_results"), **options)
 
 def update(public_id, **options):
     resource_type = options.pop("resource_type", "image")
@@ -80,6 +80,7 @@ def update(public_id, **options):
     upload_options = only(options, "moderation_status", "raw_convert", "ocr", "categorization", "detection", "similarity_search", "background_removal")
     if "tags" in options: upload_options["tags"] = ",".join(utils.build_array(options["tags"]))
     if "face_coordinates" in options: upload_options["face_coordinates"] = utils.encode_double_array(options.get("face_coordinates")) 
+    if "custom_coordinates" in options: upload_options["custom_coordinates"] = utils.encode_double_array(options.get("custom_coordinates")) 
     if "context" in options: upload_options["context"] = utils.encode_dict(options.get("context")) 
     if "auto_tagging" in options: upload_options["auto_tagging"] = float(options.get("auto_tagging"))
     return call_api("post", uri, upload_options, **options)
@@ -89,26 +90,26 @@ def delete_resources(public_ids, **options):
     type = options.pop("type", "upload")
     uri = ["resources", resource_type, type]
     params = [("public_ids[]", public_id) for public_id in public_ids]
-    optional = list(only(options, "keep_original", "next_cursor").items())
+    optional = list(only(options, "keep_original", "next_cursor", "invalidate").items())
     return call_api("delete", uri, params + optional, **options)
 
 def delete_resources_by_prefix(prefix, **options):
     resource_type = options.pop("resource_type", "image")
     type = options.pop("type", "upload")
     uri = ["resources", resource_type, type]
-    return call_api("delete", uri, dict(only(options, "keep_original", "next_cursor"), prefix=prefix), **options)
+    return call_api("delete", uri, dict(only(options, "keep_original", "next_cursor", "invalidate"), prefix=prefix), **options)
 
 def delete_all_resources(**options):
     resource_type = options.pop("resource_type", "image")
     type = options.pop("type", "upload")
     uri = ["resources", resource_type, type]
     optional = list(only(options, "keep_original", "next_cursor").items())
-    return call_api("delete", uri, dict(only(options, "keep_original", "next_cursor"), all=True), **options)
+    return call_api("delete", uri, dict(only(options, "keep_original", "next_cursor", "invalidate"), all=True), **options)
 
 def delete_resources_by_tag(tag, **options):
     resource_type = options.pop("resource_type", "image")
     uri = ["resources", resource_type, "tags", tag]
-    return call_api("delete", uri, only(options, "keep_original", "next_cursor"), **options)
+    return call_api("delete", uri, only(options, "keep_original", "next_cursor", "invalidate"), **options)
 
 def delete_derived_resources(derived_resource_ids, **options):
     uri = ["derived_resources"]
@@ -173,6 +174,12 @@ def create_upload_preset(**options):
     params.update(only(options, "unsigned", "disallow_public_id", "name"))
     return call_api("post", uri, params, **options)
 
+def root_folders(**options):
+    return call_api("get", ["folders"], {}, **options)
+
+def subfolders(of_folder_path, **options):
+    return call_api("get", ["folders", of_folder_path], {}, **options)
+
 def call_api(method, uri, params, **options):
     prefix = options.pop("upload_prefix", cloudinary.config().upload_prefix) or "https://api.cloudinary.com"
     cloud_name = options.pop("cloud_name", cloudinary.config().cloud_name)
