@@ -21,7 +21,7 @@ def unsigned_upload(file, upload_preset, **options):
 def upload_image(file, **options):
     result = upload(file, **options)
     return cloudinary.CloudinaryImage(result["public_id"], version=str(result["version"]),
-        format=result["format"], metadata=result)
+        format=result.get("format"), metadata=result)
 
 def upload_large(file, **options):
     """ Upload large raw files. Note that public_id should include an extension for best results. """
@@ -35,7 +35,7 @@ def upload_large(file, **options):
             chunk_io.name = basename(file)
             chunk = file_io.read(20000000)
             upload = upload_large_part(chunk_io, public_id=public_id,
-                            upload_id=upload_id, part_number=index, final=chunk != "", **options)
+                            upload_id=upload_id, part_number=index, final=chunk == "", **options)
             upload_id = upload.get("upload_id")
             public_id = upload.get("public_id")
             index += 1
@@ -50,6 +50,7 @@ def upload_large_part(file, **options):
         "final": options.get("final"),
         "part_number": options.get("part_number"),
         "upload_id": options.get("upload_id"),
+        "tags": options.get("tags") and ",".join(utils.build_array(options["tags"])),
         "public_id": options.get("public_id")
     }
     return call_api("upload_large", params, resource_type="raw", file=file, **options)
@@ -83,7 +84,8 @@ def explicit(public_id, **options):
         "headers": utils.build_custom_headers(options.get("headers")),
         "eager": utils.build_eager(options.get("eager")),
         "tags": options.get("tags") and ",".join(utils.build_array(options["tags"])),
-        "face_coordinates": utils.encode_double_array(options.get("face_coordinates"))}
+        "face_coordinates": utils.encode_double_array(options.get("face_coordinates")),
+        "custom_coordinates": utils.encode_double_array(options.get("custom_coordinates"))}
      return call_api("explicit", params, **options)
 
 def generate_sprite(tag, **options):
@@ -171,7 +173,7 @@ def call_api(action, params, **options):
             # Register the streaming http handlers with urllib2
             register_openers()
     
-        datagen = to_bytes("")
+        datagen = []
         headers = {}
         if "file" in options:
             file = options["file"]
@@ -182,6 +184,11 @@ def call_api(action, params, **options):
                 datagen, headers = multipart_encode({'file': file_io})
             else:
                 param_list.append(("file", file))
+
+        if _is_gae():
+            # Might not be needed in the future but for now this is needed in GAE
+            datagen = "".join(datagen)
+
         request = urllib2.Request(api_url + "?" + urlencode(param_list), datagen, headers)
         request.add_header("User-Agent", cloudinary.USER_AGENT)
     
@@ -218,4 +225,7 @@ def call_api(action, params, **options):
         return result
     finally:
         if file_io: file_io.close()    
-    
+
+def _is_gae():
+   import httplib
+   return 'appengine' in str(httplib.HTTP)
